"""Auto-generated model profiles.

DO NOT EDIT THIS FILE MANUALLY.
This file is generated by the langchain-profiles CLI tool.

It contains data derived from the models.dev project.

Source: https://github.com/sst/models.dev
License: MIT License

To update these data, refer to the instructions here:

https://docs.langchain.com/oss/python/langchain/models#updating-or-overwriting-profile-data
"""

from typing import Any

_PROFILES: dict[str, dict[str, Any]] = {
    "MiniMaxAI/MiniMax-M2.1": {
        "max_input_tokens": 204800,
        "max_output_tokens": 131072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "MiniMaxAI/MiniMax-M2.5": {
        "max_input_tokens": 204800,
        "max_output_tokens": 131072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "Qwen/Qwen3-235B-A22B-Thinking-2507": {
        "max_input_tokens": 262144,
        "max_output_tokens": 131072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "Qwen/Qwen3-Coder-480B-A35B-Instruct": {
        "max_input_tokens": 262144,
        "max_output_tokens": 66536,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
    },
    "Qwen/Qwen3-Coder-Next": {
        "max_input_tokens": 262144,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
    },
    "Qwen/Qwen3-Embedding-4B": {
        "max_input_tokens": 32000,
        "max_output_tokens": 2048,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
    },
    "Qwen/Qwen3-Embedding-8B": {
        "max_input_tokens": 32000,
        "max_output_tokens": 4096,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": False,
    },
    "Qwen/Qwen3-Next-80B-A3B-Instruct": {
        "max_input_tokens": 262144,
        "max_output_tokens": 66536,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
    },
    "Qwen/Qwen3-Next-80B-A3B-Thinking": {
        "max_input_tokens": 262144,
        "max_output_tokens": 131072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
    },
    "Qwen/Qwen3.5-397B-A17B": {
        "max_input_tokens": 262144,
        "max_output_tokens": 32768,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "XiaomiMiMo/MiMo-V2-Flash": {
        "max_input_tokens": 262144,
        "max_output_tokens": 4096,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "deepseek-ai/DeepSeek-R1-0528": {
        "max_input_tokens": 163840,
        "max_output_tokens": 163840,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "deepseek-ai/DeepSeek-V3.2": {
        "max_input_tokens": 163840,
        "max_output_tokens": 65536,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "moonshotai/Kimi-K2-Instruct": {
        "max_input_tokens": 131072,
        "max_output_tokens": 16384,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
    },
    "moonshotai/Kimi-K2-Instruct-0905": {
        "max_input_tokens": 262144,
        "max_output_tokens": 16384,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": False,
        "tool_calling": True,
    },
    "moonshotai/Kimi-K2-Thinking": {
        "max_input_tokens": 262144,
        "max_output_tokens": 262144,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "moonshotai/Kimi-K2.5": {
        "max_input_tokens": 262144,
        "max_output_tokens": 262144,
        "text_inputs": True,
        "image_inputs": True,
        "audio_inputs": False,
        "video_inputs": True,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "zai-org/GLM-4.7": {
        "max_input_tokens": 204800,
        "max_output_tokens": 131072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "zai-org/GLM-4.7-Flash": {
        "max_input_tokens": 200000,
        "max_output_tokens": 128000,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
    "zai-org/GLM-5": {
        "max_input_tokens": 202752,
        "max_output_tokens": 131072,
        "text_inputs": True,
        "image_inputs": False,
        "audio_inputs": False,
        "video_inputs": False,
        "text_outputs": True,
        "image_outputs": False,
        "audio_outputs": False,
        "video_outputs": False,
        "reasoning_output": True,
        "tool_calling": True,
    },
}
